--
-- Copyright (C) 2012-2016 Snail <https://github.com/snail23/snailui/>
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the
-- Free Software Foundation; either version 2 of the License, or (at your
-- option) any later version.
--
-- This program is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
-- FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
-- more details.
--
-- You should have received a copy of the GNU General Public License along
-- with this program. If not, see <http://www.gnu.org/licenses/>.
--

local _
local Index
local Nameplates

local function HandleNameplate(Self)
	local Frames =
	{
		"AbsorbBar",
		"AggroWarningTexture",
		"Border",
		"CastBar",
		"EliteIcon",
		"HealthBar",
		"HighLevelIcon",
		"Highlight",
		"LevelText",
		"RaidTargetIcon"
	}

	for I = 1, #Frames do
		Self.ArtContainer[Frames[I]]:Hide()
		Self.ArtContainer[Frames[I]]:SetParent(Nameplates.Trash)
	end

	local function UpdateNameplate(Self)
		Self:SetScale(0.75)
		Self.Health:SetValue(Self.ArtContainer.HealthBar:GetValue())

		if Self.ArtContainer.AggroWarningTexture:IsShown() then
			local AR, AG, AB = Self.ArtContainer.AggroWarningTexture:GetVertexColor()

			if (AR > 0.9) and (AG == 0) and (AB == 0) then
				Self.Background:SetTexture(1, 0, 0.75)
				Self.Health:SetStatusBarColor(1, 0, 0.75)
				Self.Health.Name:SetTextColor(Configuration.TextColorPercentage, 0, 0.75 * Configuration.TextColorPercentage)
			else 
				Self.Background:SetTexture(AR, AG, AB)
				Self.Health:SetStatusBarColor(AR, AG, AB)
				Self.Health.Name:SetTextColor(AR * Configuration.TextColorPercentage, AG * Configuration.TextColorPercentage, AB * Configuration.TextColorPercentage)
			end
		else
			local R, G, B = Self.ArtContainer.HealthBar:GetStatusBarColor()

			Self.Background:SetTexture(R, G, B)
			Self.Health:SetStatusBarColor(R, G, B)
			Self.Health.Name:SetTextColor(R * Configuration.TextColorPercentage, G * Configuration.TextColorPercentage, B * Configuration.TextColorPercentage)
		end

		Self.Health.Name:SetText(Self.NameContainer.NameText:GetText())
	end

	Self:SetHeight(GetConfiguration().Nameplates.Height)
	Self:SetScale(0.75)
	Self:SetScript("OnUpdate", UpdateNameplate)
	Self:SetScript("OnShow",
		function(Self)
			Self.Health:SetMinMaxValues(Self.ArtContainer.HealthBar:GetMinMaxValues())
			UpdateNameplate(Self)
		end
	)

	Self.ArtContainer:Hide()
	Self.ArtContainer:SetParent(Nameplates.Trash)

	Self.ArtContainer.HealthBar:SetScript("OnValueChanged",
		function(_, Value)
			Self.Health:SetValue(Value)
		end
	)

	Self.Background = Self:CreateTexture(nil, "LOW")
	Self.Background:SetPoint("CENTER")
	Self.Background:SetSize(math.floor(Self:GetWidth() + 0.5) - 2, GetConfiguration().Nameplates.Height - 2)

	Self.Border = Self:CreateTexture(nil, "BACKGROUND")
	Self.Border:SetPoint("CENTER")
	Self.Border:SetSize(math.floor(Self:GetWidth() + 0.5), GetConfiguration().Nameplates.Height)
	Self.Border:SetTexture(0, 0, 0)

	Self.Health = CreateFrame("StatusBar", nil, Self)
	Self.Health:SetOrientation(GetConfiguration().Nameplates.HealthBar.Orientation)
	Self.Health:SetPoint(GetConfiguration().Nameplates.HealthBar.Anchor, GetConfiguration().Nameplates.HealthBar.X, GetConfiguration().Nameplates.HealthBar.Y)
	Self.Health:SetSize(math.floor(Self:GetWidth() + 0.5) - 6, GetConfiguration().Nameplates.HealthBar.Height - 2)
	Self.Health:SetStatusBarTexture(Configuration.Texture)
	
	SmoothBar(Self, Self.Health)

	Self.Health.Border = Self.Health:CreateTexture(nil, "LOW")
	Self.Health.Border:SetPoint("CENTER")
	Self.Health.Border:SetSize(math.floor(Self:GetWidth() + 0.5) - 4, GetConfiguration().Nameplates.HealthBar.Height)
	Self.Health.Border:SetTexture(0, 0, 0)

	if GetConfiguration().Nameplates.HealthBar.Name then
		Self.Health.Name = Self.Health:CreateFontString(nil, "OVERLAY")
		Self.Health.Name:SetFont(Configuration.Font.Name, Configuration.Font.Size, Configuration.Font.Outline)
		Self.Health.Name:SetPoint(GetConfiguration().Nameplates.HealthBar.Name.Anchor, GetConfiguration().Nameplates.HealthBar.Name.X, GetConfiguration().Nameplates.HealthBar.Name.Y)
	end

	Self.NameContainer:Hide()
	Self.NameContainer:SetParent(Nameplates.Trash)

	Self.Skinned = true
	Index = Index + 1
end

function HandleNameplates()
	if GetConfiguration().Nameplates then
		Nameplates = CreateFrame("Frame", nil, UIParent)
		Nameplates.Trash = CreateFrame("Frame")
		Nameplates.Trash:Hide()

		Nameplates:SetScript("OnUpdate",
			function(Self, ElapsedTime)
				local Nameplate

				if not Index then
					for _, Nameplate in next, {WorldFrame:GetChildren()} do
						local Name = Nameplate:GetName()

						if Name and string.match(Name, "^NamePlate%d+$") and not Nameplate.Skinned then
							Index = string.match(Name, "(%d+)")
							HandleNameplate(Nameplate)
						end
					end
				else
					Nameplate = _G["NamePlate" .. Index]

					if Nameplate and not Nameplate.Skinned then
						HandleNameplate(Nameplate)
					end
				end
			end
		)
	end
end
